@@ -25,7 +25,7 @@ SEXP ffirst_impl(SEXP x, int ng, SEXP g, int narm, int *gl) {
           break;
         }
         case STRSXP: {
-          SEXP *px = STRING_PTR(x);
+          const SEXP *px = SEXPPTR_RO(x);
           while(px[j] == NA_STRING && j != end) ++j;
           SET_STRING_ELT(out, 0, px[j]);
           break;
@@ -38,7 +38,7 @@ SEXP ffirst_impl(SEXP x, int ng, SEXP g, int narm, int *gl) {
           break;
         }
         case VECSXP: {
-          SEXP *px = SEXPPTR(x);
+          const SEXP *px = SEXPPTR_RO(x);
           while(length(px[j]) == 0 && j != end) ++j;
           SET_VECTOR_ELT(out, 0, px[j]);
           break;
@@ -86,7 +86,8 @@ SEXP ffirst_impl(SEXP x, int ng, SEXP g, int narm, int *gl) {
         break;
       }
       case STRSXP: {
-        SEXP *px = STRING_PTR(x), *pout = STRING_PTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng; i--; ) pout[i] = NA_STRING;
         --pout;
         for(int i = 0; i != l; ++i) {
@@ -115,7 +116,8 @@ SEXP ffirst_impl(SEXP x, int ng, SEXP g, int narm, int *gl) {
         break;
       }
       case VECSXP: {
-        SEXP *px = SEXPPTR(x), *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng; i--; ) pout[i] = R_NilValue; // R_NilValue or just leave empty ??
         --pout;
         for(int i = 0; i != l; ++i) {
@@ -155,12 +157,14 @@ SEXP ffirst_impl(SEXP x, int ng, SEXP g, int narm, int *gl) {
         break;
       }
       case STRSXP:{
-        SEXP *px = STRING_PTR(x)-1, *pout = STRING_PTR(out);
+        const SEXP *px = SEXPPTR_RO(x)-1;
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng; i--; ) pout[i] = gl[i] == NA_INTEGER ? NA_STRING : px[gl[i]];
         break;
       }
       case VECSXP: {
-        SEXP *px = SEXPPTR(x)-1, *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x)-1;
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng; i--; ) pout[i] = gl[i] == NA_INTEGER ? R_NilValue : px[gl[i]];
         break;
       }
@@ -213,7 +217,7 @@ SEXP ffirstlC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm) {
   int l = length(x), *pgl, ng = asInteger(Rng), narm = asLogical(Rnarm), nprotect = 1;
   if(ng > 0 && !narm) {
     if(length(gst) != ng) {
-    // Cant use integer array here because apparently it is removed by the garbage collector when passed to a new function
+    // Can't use integer array here because apparently it is removed by the garbage collector when passed to a new function
     SEXP gl = PROTECT(allocVector(INTSXP, ng)); ++nprotect;
     int *pg = INTEGER(g), lg = length(g); // gl[ng],
     pgl = INTEGER(gl); // pgl = &gl[0];
@@ -225,8 +229,8 @@ SEXP ffirstlC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm) {
   } else pgl = &l; // To avoid Wmaybe uninitialized..
   // return ffirst_impl(VECTOR_ELT(x, 0), ng, g, narm, pgl);
   SEXP out = PROTECT(allocVector(VECSXP, l));
-  SEXP *px = SEXPPTR(x), *pout = SEXPPTR(out);
-  for(int j = 0; j != l; ++j) pout[j] = ffirst_impl(px[j], ng, g, narm, pgl);
+  const SEXP *px = SEXPPTR_RO(x);
+  for(int j = 0; j != l; ++j) SET_VECTOR_ELT(out, j, ffirst_impl(px[j], ng, g, narm, pgl));
   DFcopyAttr(out, x, ng);
   UNPROTECT(nprotect);
   return out;
@@ -252,7 +256,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case STRSXP: {
-        SEXP *px = STRING_PTR(x), *pout = STRING_PTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int j = 0, i = 0; j != col; ++j) {
           while(px[i] == NA_STRING && i != end) ++i;
           pout[j] = px[i]; px += l; i = 0;
@@ -269,7 +274,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case VECSXP: {
-        SEXP *px = SEXPPTR(x), *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int j = 0, i = 0; j != col; ++j) {
           while(length(px[i]) == 0 && i != end) ++i;
           pout[j] = px[i]; px += l; i = 0;
@@ -293,7 +299,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
       }
       case STRSXP:
       case VECSXP: {
-        SEXP *px = SEXPPTR(x), *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int j = 0; j != col; ++j) pout[j] = px[j * l];
         break;
       }
@@ -321,7 +328,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case STRSXP: {
-        SEXP *px = STRING_PTR(x), *pout = STRING_PTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng * col; i--; ) pout[i] = NA_STRING;
         --pout;
         for(int j = 0; j != col; ++j) {
@@ -342,7 +350,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case VECSXP: {
-        SEXP *px = SEXPPTR(x), *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x);
+        SEXP *pout = SEXPPTR(out);
         for(int i = ng * col; i--; ) pout[i] = R_NilValue;
         --pout;
         for(int j = 0; j != col; ++j) {
@@ -383,7 +392,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case STRSXP: {
-        SEXP *px = STRING_PTR(x)-1, *pout = STRING_PTR(out);
+        const SEXP *px = SEXPPTR_RO(x)-1;
+        SEXP *pout = SEXPPTR(out);
         for(int j = 0; j != col; ++j) {
           for(int i = ng; i--; ) pout[i] = pgl[i] == NA_INTEGER ? NA_STRING : px[pgl[i]];
           px += l; pout += ng;
@@ -391,7 +401,8 @@ SEXP ffirstmC(SEXP x, SEXP Rng, SEXP g, SEXP gst, SEXP Rnarm, SEXP Rdrop) {
         break;
       }
       case VECSXP: {
-        SEXP *px = SEXPPTR(x)-1, *pout = SEXPPTR(out);
+        const SEXP *px = SEXPPTR_RO(x)-1;
+        SEXP *pout = SEXPPTR(out);
         for(int j = 0; j != col; ++j) {
           for(int i = ng; i--; ) pout[i] = pgl[i] == NA_INTEGER ? R_NilValue : px[pgl[i]];
           px += l; pout += ng;
